/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993 by Mark Wahl
   For license details see documentation
   Procedures for connect (connect.c)
   created: Mark Wahl DL4YBG 94/03/12
   updated: Mark Wahl DL4YBG 94/03/21
*/

#include "tnt.h"
#include "connect.h"

extern struct channel_stat ch_stat[CHANNELS];
extern char name_file_name[];
extern char tnt_dir[];
extern int tnc_command;

char connect_str[] = "onnected to ";

static char xc_active_text[] = "XCONNECT already active";
static char xc_notfound_text[] = "Routing file not found";
static char xc_corrupt_text[] = "Routing file corrupt";
static char xc_maxdigis_text[] = "Digipeater-table full";
static char xc_maxalloc_text[] = "Too many recursions";
static char xc_noroute_text[] = "Destination not known";
static char xc_abort_text[] = "XCONNECT aborted";
static char xc_noxconn_text[] = "XCONNECT not active";

static struct xconn_stat xc_stat[CHANNELS];

static int test_connected(channel,buffer)
int channel;
char *buffer;
{
  char str_buf[257];
  int len;
  char *str_ptr;
  int result;
  
  
  /* copy buffer to string, \0 may occure in buffer, but if connected
     is assumed to be received, plain ASCII will be received */
  result = 0;
  len = (int)(*buffer + 1);
  memcpy(str_buf,buffer+1,len);
  str_buf[len] = '\0';
  /* try to find connect-string */
  if ((str_ptr = strstr(str_buf,connect_str)) == NULL) return(T_NOCON);
  /* check if connected */
  if (str_ptr > (str_buf + 2)) {
    if ((tolower(*(str_ptr - 1)) == 'c') && ((*(str_ptr - 2)) == ' '))
      result = T_CON;
  }
  /* else check if reconnected */
  if (!result && (str_ptr > (str_buf + 3))) {
    if (((*(str_ptr - 1)) == 'c') && ((*(str_ptr - 2)) == 'e') &&
        (tolower(*(str_ptr - 3)) == 'r'))
      result = T_RECON;
  }
  if (!result) return(result);
  str_ptr += strlen(connect_str);
  /* now get new callsign */
  if (sscanf(str_ptr,"%30s",ch_stat[channel].disp_call) != 1) return(0);
  stat_display(channel);
  if (result == T_CON) /* if connect copy to call for logbook */
    strcpy(ch_stat[channel].log_call,ch_stat[channel].disp_call);
  return(result);
}

#define LINELENGTH 127

static int route_call(channel,len,mode,str,nextlink,callflag)
int channel;
int len;
int mode;
char *str;
int nextlink;
int callflag;
{  
  int found;
  int corrupt;
  int num;
  char *line_ptr;
  FILE *xc_fp;
  char *call_ptr;
  char *tmp_ptr;
  char tmpstr[160];

  strcpy(tmpstr,tnt_dir);
  strcat(tmpstr,name_file_name);
  if ((xc_fp = fopen(tmpstr,"r")) == NULL) {
    cmd_display(mode,channel,xc_notfound_text,1);
    return(0);
  }
  found = 0;
  line_ptr = malloc(LINELENGTH + 1);
  while (!found) {
    corrupt = 1;
    if ((fgets(line_ptr,LINELENGTH,xc_fp)) != NULL) {
      if (*(line_ptr + strlen(line_ptr) -1) == '\n') {
        if ((call_ptr = strtok(line_ptr," \t")) != NULL ) {
          corrupt = 0;
          tmp_ptr = callflag ? call_ptr : (call_ptr + 2);
          if (strcmp(tmp_ptr,str) == 0) {
            found = 1;
          }
        }
      }
    }
    else {
      return(C_NOTFOUND);
    }
    if (corrupt) {
      cmd_display(mode,channel,xc_corrupt_text,1);
      free(line_ptr);
      fclose(xc_fp);
      return(0);
    }
  }
  tmp_ptr = strtok(NULL,";"); 
  if ((tmp_ptr == NULL) || (*(tmp_ptr + strlen(tmp_ptr) - 1) == '\n')) {
    /* station can be reached directly */
    free(line_ptr);
    fclose(xc_fp);
    return(C_DIRECT);  
  }
  /* station has routing data available */
  num = xc_stat[channel].num_digi;
  while ((xc_stat[channel].digiinfo[num].nextdigi =
         strtok(NULL," \t")) != NULL) {
    tmp_ptr = xc_stat[channel].digiinfo[num].nextdigi;
    if (*(tmp_ptr = tmp_ptr + strlen(tmp_ptr) - 1) == '\n')
      *tmp_ptr = '\0';
    num++;
    if (num >= MAXDIGIS) {
      cmd_display(mode,channel,xc_maxdigis_text,1);
      free(line_ptr);
      fclose(xc_fp);
      return(0);
    }
    xc_stat[channel].digiinfo[num - 1].nextlink = num;
  }
  if (num == xc_stat[channel].num_digi) {
    /* no routing data found after ; */
    cmd_display(mode,channel,xc_corrupt_text,1);
    free(line_ptr);
    fclose(xc_fp);
    return(0);
  }
  xc_stat[channel].digiinfo[num - 1].nextlink = nextlink;
  xc_stat[channel].num_digi = num;
  xc_stat[channel].alloc[xc_stat[channel].num_alloc] = line_ptr;
  xc_stat[channel].num_alloc++;
  fclose(xc_fp);
  return(C_ROUTE);
}

static int gen_con_string(channel,str,ctype)
int channel;
char *str;
int ctype;
{
  char tmp_str[256];
  char *tmp_ptr;
  int nodetype;
  int digi;
  
  strcpy(tmp_str,"");
  digi = xc_stat[channel].cur_digi;
  while (1) {
    tmp_ptr = xc_stat[channel].digiinfo[digi].nextdigi;
    digi = xc_stat[channel].digiinfo[digi].nextlink;
    xc_stat[channel].cur_digi = digi;
    if (digi == D_EOL) {
      nodetype = CT_END;
      break;
    }
    if (*tmp_ptr == 'N') {
      nodetype = CT_NODE;
      break;
    }
    if (*tmp_ptr == 'T') {
      nodetype = CT_USER;
      break;
    }
    strcat(tmp_str," "); 
    strcat(tmp_str,tmp_ptr+2);
  }
  switch (ctype) {
  case CT_CONNECT:
    strcpy(str,"C");
    break;
  case CT_NODE:
    strcpy(str,"C ");
    break;
  case CT_USER:
    strcpy(str,"//C ");
    break;
  }
  strcat(str,tmp_ptr+2);
  strcat(str,tmp_str);
  if ((ctype == CT_NODE) || (ctype == CT_USER)) {
    strcat(str,"\r");
  }
  return(nodetype);
}

static void free_alloc(channel)
int channel;
{
  int i;
  
  if (xc_stat[channel].num_alloc) {
    for (i = 0; i < xc_stat[channel].num_alloc; i++) {
      free(xc_stat[channel].alloc[i]);
    }
  }
}

void close_xconnect(channel)
int channel;
{
  if (xc_stat[channel].active) {
    xc_stat[channel].active = 0;
    statlin_update();
    free_alloc(channel);
  }
}

void cmd_xconnect(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int result;
  char tnc_string[256];
  int i;
  int first_digi;
  int second_digi;
  int digi;
  int expand_end;
  
  for (i = 0; i < len; i++) {
    str[i] = toupper(str[i]);
  }
  if (xc_stat[channel].active) {
    if (strcmp(str,"OFF") == 0) {
      close_xconnect(channel);
      cmd_display(mode,channel,xc_abort_text,1);
      return;
    }
    else {
      cmd_display(mode,channel,xc_active_text,1);
      return;
    }
  }
  if (strcmp(str,"OFF") == 0) {
    cmd_display(mode,channel,xc_noxconn_text,1);
    return;
  }
  xc_stat[channel].num_digi = 0;
  xc_stat[channel].num_alloc = 0;
  xc_stat[channel].first_digi = 0;
  if ((result = route_call(channel,len,mode,str,D_EOL,0)) == 0) return;
  switch (result) {
  case C_NOTFOUND:
    cmd_display(mode,channel,xc_noroute_text,1);
    return;
  case C_DIRECT:
    tnc_command = 1;
    strcpy(tnc_string,"C");
    strcat(tnc_string,str);
    queue_cmd_data(channel,X_COMM,strlen(tnc_string),mode,tnc_string);
    break;
  case C_ROUTE:
    expand_end = 0;
    while (!expand_end) {
      first_digi = xc_stat[channel].first_digi;
      second_digi = xc_stat[channel].digiinfo[first_digi].nextlink;
      xc_stat[channel].first_digi = xc_stat[channel].num_digi;
      if ((result = route_call(channel,len,mode,
                    xc_stat[channel].digiinfo[first_digi].nextdigi,
                    second_digi,1)) == 0) return;
      switch (result) {
      case C_NOTFOUND:
      case C_DIRECT:
        xc_stat[channel].first_digi = first_digi;
        expand_end = 1;
        break;
      case C_ROUTE:
        break;
      }
    }
    xc_stat[channel].cur_digi = xc_stat[channel].first_digi;
    xc_stat[channel].ctype = gen_con_string(channel,tnc_string,CT_CONNECT);
    if (xc_stat[channel].ctype == CT_END) {
      free_alloc(channel);
    }
    else {
     xc_stat[channel].active = 1;
     statlin_update();
    }
    tnc_command = 1;
    queue_cmd_data(channel,X_COMM,strlen(tnc_string),mode,tnc_string);
    break;
  }
}

void next_connect(channel)
int channel;
{
  char tnc_string[256];
  int flag;

  if (xc_stat[channel].active) {
    xc_stat[channel].ctype = gen_con_string(channel,tnc_string,
                                            xc_stat[channel].ctype);
    if (xc_stat[channel].ctype == CT_END) {
      close_xconnect(channel);
    }
    flag = 0;
    rem_data_display(channel,tnc_string);
    queue_cmd_data(channel,X_DATA,strlen(tnc_string),flag,tnc_string);
  }
}

void connect_update(channel,buffer)
int channel;
char *buffer;
{
  switch (test_connected(channel,buffer)) {
  case T_CON:
    next_connect(channel);
    break;
  case T_RECON:
    close_xconnect(channel);
    break;
  default:
    break;
  }
}

int xconnect_active(channel)
int channel;
{
  return(xc_stat[channel].active > 0);
}
