/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993 by Mark Wahl
   For license details see documentation
   Procedures for display (display.c)
   created: Mark Wahl DL4YBG 93/08/01
   updated: Mark Wahl DL4YBG 94/03/21
*/

#include "tnt.h"

extern int act_channel;
extern int act_mode;
extern struct channel_stat ch_stat[CHANNELS];
extern char signon[];
extern int chan_flag;
extern int pause_flag[CHANNELS];
extern int resync;
extern char att_normal;
extern char att_statline;
extern char att_monitor;
extern char att_cstatline;
extern char att_controlchar; 
extern char att_remote;
extern int lines_command;
extern int lines_monitor;
extern int lines_input;
extern int lines_output;
extern char rem_tnt_str[];
extern char rem_newlin_str[];
extern int infobell_flag;
extern int cbell_flag;
extern int LINES,COLS;
extern int tnc_busy;
extern struct tx_file tx_file_com;

static struct window prewin[CHANNELS];
static struct window statwin[CHANNELS];
static struct window textwin[CHANNELS];
static struct window cmdwin;
static struct window monwin;
static struct window statlin;

static struct real_layout layout[CHANNELS+2][5];

static int old_mode;

static char *ax_state[] = {
  "************",
  " LINK SETUP ",
  " FRAME REJ *",
  " DISC. LINK ",
  " INFO XFER *",
  " REJECT SND ",
  " WAIT ACK **",
  " LOCAL BUSY ",
  " REM. BUSY *",
  " BOTH BUSY *",
  " LOCAL BUSY ",
  " REM. BUSY *",
  " BOTH BUSY *",
  " LOCAL BUSY ",
  " REM. BUSY *",
  " BOTH BUSY *"
};
  
void statlin_update()
{
  int i;
  char c;
  char tmp;
  char tmpstr[NUM_COLUMNS+1];
  
  switch (act_mode) {
  case M_CONNECT:
    strcpy(tmpstr,"\015*** CONNECT *");
    if (resync)
      tmp = 'S';
    else if (tnc_busy)
      tmp = 'B';
    else if (pause_flag[act_channel])
      tmp = 'P';
    else if (xconnect_active(act_channel))
      tmp = 'X';
    else if (insert_active())
      tmp = 'I';
    else
      tmp = '*';
    break;
      strncat(tmpstr,&tmp,1);
  case M_COMMAND:
    strcpy(tmpstr,"\015*** COMMAND *");
    if (resync)
      tmp = 'S';
    else if (tnc_busy)
      tmp = 'B';
    else if (insert_active())
      tmp = 'I';
    else 
      tmp = '*';
    break;
  case M_MONITOR:
    strcpy(tmpstr,"\015*** MONITOR *");
    if (resync)
      tmp = 'S';
    else if (tnc_busy)
      tmp = 'B';
    else if (pause_flag[0])
      tmp = 'P';
    else 
      tmp = '*';
    break;  
  }
  strncat(tmpstr,&tmp,1);
  strcat(tmpstr,"* CHANNEL ");
  if (chan_flag) {
    tmp = '?';
  }
  else {
    if (act_channel > 9) {
      tmp = act_channel-10+'@';
    }
    else {
      tmp = act_channel+'0';
    }
  }
  strncat(tmpstr,&tmp,1);
  strcat(tmpstr," *");
  show_rxfile(act_channel,tmpstr);
  tmp = '*';
  strncat(tmpstr,&tmp,1);
  show_txfile(act_channel,tmpstr);
  strcat(tmpstr,"* ");
  for (i = 0; i < CHANNELS; i++) {
    if ((i == act_channel) && (act_mode == M_CONNECT)) {
      ch_stat[i].not_disp = 0;
    }
    if (ch_stat[i].not_disp) {
      c = '0';
      if (i > 9) c = '@' - 10;
      tmp = i+c;
    }
    else if (ch_stat[i].state) {
      tmp = '+';
    }
    else {
      tmp = SPACE;
    }
    strncat(tmpstr,&tmp,1);
  }
  strcat(tmpstr," **");
#ifdef TNT_CHAMBER
  for (i=0;i<strlen(tmpstr);i++) {
    if (tmpstr[i] == '*')
      tmpstr[i] = '-';
  }
#endif
  win_stringout(tmpstr,&statlin,0);
}

void newdata(channel)
int channel;
{
  ch_stat[channel].not_disp = 1;
  statlin_update();
}

void sel_monitor()
{
  if (act_mode == M_MONITOR) {
    if (old_mode == M_COMMAND) {
      real_screen(&layout[COMMAND][0]);
    }
    else {
      real_screen(&layout[act_channel][0]);
    }
    act_mode = old_mode;
  }
  else {
    old_mode = act_mode;
    real_screen(&layout[MONITOR][0]);
    act_mode = M_MONITOR;
  }
  statlin_update();
}

void sel_command()
{
  if (act_mode == M_COMMAND) return;
  real_screen(&layout[COMMAND][0]);
  act_mode = M_COMMAND;
  statlin_update();
}

void sel_connect(channel)
int channel;
{
  if ((channel == act_channel) && (act_mode == M_CONNECT)) return;  
  real_screen(&layout[channel][0]);
  act_mode = M_CONNECT;
  act_channel = channel;
  statlin_update();
}

void stat_display(channel)
{
  int i;
  int stat_len;
  int snd_frms;
  char tmp;
  char stat_str[81];
  
  strcpy(stat_str,"");
  strcat(stat_str,"\015**");
  if (channel == 0) {
    strcat(stat_str,"************************************* ");
    strcat(stat_str,"UNPROTO CHANNEL ************************");
  }
  else {
    strcat(stat_str,ax_state[ch_stat[channel].state & 0x0f]);
    strcat(stat_str,"**");
    if (ch_stat[channel].snd_frms) {
      strcat(stat_str," S=");
      snd_frms = ch_stat[channel].snd_frms + ch_stat[channel].snd_queue_frms;
      if (snd_frms > 9) {
        tmp = snd_frms/10+'0';
        strncat(stat_str,&tmp,1);
        tmp = snd_frms%10+'0';
        strncat(stat_str,&tmp,1);
        tmp = SPACE;
        strncat(stat_str,&tmp,1);
      }
      else {
        tmp = snd_frms+'0';
        strncat(stat_str,&tmp,1);
        strcat(stat_str," *");
      }
    }
    else strcat(stat_str,"******");
    strcat(stat_str,"**");
    if (ch_stat[channel].unacked) {
      strcat(stat_str," U=");
      tmp = ch_stat[channel].unacked+'0';
      strncat(stat_str,&tmp,1);
      tmp = SPACE;
      strncat(stat_str,&tmp,1);
    }
    else strcat(stat_str,"*****");
    strcat(stat_str,"**");
    if (ch_stat[channel].tries) {
      strcat(stat_str," R=");
      if (ch_stat[channel].tries > 9) {
        tmp = ch_stat[channel].tries/10+'0';
        strncat(stat_str,&tmp,1);
        tmp = ch_stat[channel].tries%10+'0';
        strncat(stat_str,&tmp,1);
        tmp = SPACE;
        strncat(stat_str,&tmp,1);
      }
      else {
        tmp = ch_stat[channel].tries+'0';
        strncat(stat_str,&tmp,1);
        strcat(stat_str," *");
      }
    }
    else strcat(stat_str,"******");
    strcat(stat_str,"** ");

    if (ch_stat[channel].disp_call[0] == '\0')   
      strncat(stat_str,ch_stat[channel].call,40);
    else
      strncat(stat_str,ch_stat[channel].disp_call,40);
    if ((stat_len = strlen(stat_str)) < 81) {
      strncat(stat_str,
              " ***************************************",81 - stat_len);
    }
  }
#ifdef TNT_CHAMBER
  for (i=0;i<strlen(stat_str);i++) {
    if (stat_str[i] == '*')
      stat_str[i] = '-';
  }
#endif
  win_stringout(stat_str,&statwin[channel],0);
}

void data_display(channel,buffer)
int channel;
char *buffer;
{
  char tmpstring[280];
  
  if (cbell_flag) beep();
  strcpy(tmpstring,"\r*** ");
  strcat(tmpstring,buffer);
  strcat(tmpstring,"\r");
  win_stringout(tmpstring,&textwin[channel],1);
  write_file(channel,strlen(tmpstring),RX_RCV,tmpstring);
  newdata(channel);
}

void rem_data_display(channel,buffer)
int channel;
char *buffer;
{
  win_attrib(att_remote,&textwin[channel]);
  win_stringout(buffer,&textwin[channel],1);
  win_attrib(att_normal,&textwin[channel]);
}

void rem_stat_display(channel,buffer)
int channel;
char *buffer;
{
  win_attrib(att_remote,&textwin[channel]);
  win_stringout(buffer,&textwin[channel],1);
  win_charout_cntl(C_STLINE,&textwin[channel]);
  win_attrib(att_normal,&textwin[channel]);
}

void data_display_len(channel,buffer)
int channel;
char *buffer;
{
  if (write_pty(channel,(*buffer)+1,buffer+1)) return;
  if (infobell_flag) beep();
  newdata(channel);
  if (!act_abin_rx(channel)) {
    insert_cr_tx(channel);
    win_textout(buffer,&textwin[channel],1);
    connect_update(channel,buffer);
  }
  if (!remote_input(channel,buffer))
    write_file(channel,(*buffer)+1,RX_RCV,buffer+1);
}

void rem_data_display_len(channel,buffer)
int channel;
char *buffer;
{
  win_attrib(att_remote,&textwin[channel]);
  win_textout(buffer,&textwin[channel],1);
  win_attrib(att_normal,&textwin[channel]);
}

void cmd_display(flag,channel,buffer,cr)
int flag;
int channel;
char *buffer;
int cr;
{
  int ans_flag;
  int ans_len;
  char ans_str[256];
  int con_channel;
  
  switch (flag) {
  case M_CMDSCRIPT:
    return;
  case M_CONNECT:
    if (tx_file_com.type == -1) {
      con_channel = channel;
    }
    else {
      con_channel = act_channel;
    } 
    win_stringout(buffer,&prewin[con_channel],0);
    if (cr) win_charout(CR,&prewin[con_channel],0);
    break;
  case M_REMOTE_TEMP:
  case M_REMOTE:
    strcpy(ans_str,rem_tnt_str);
    strcat(ans_str,buffer);
    strcat(ans_str,rem_newlin_str);
    ans_len = strlen(ans_str);
    ans_flag = 0;
    rem_data_display(channel,ans_str);
    queue_cmd_data(channel,X_DATA,ans_len,ans_flag,ans_str);
    break;
  default:
    win_stringout(buffer,&cmdwin,0);
    if (cr) win_charout(CR,&cmdwin,0);
    break;
  }
}
  
void moni_display(channel,buffer)
int channel;
char *buffer;
{
  win_attrib(att_monitor,&monwin);
  win_stringout(buffer,&monwin,0);
  win_attrib(att_normal,&monwin);
  win_charout(CR,&monwin,0);
}

void moni_display_len(channel,buffer)
int channel;
char *buffer;
{
  win_textout(buffer,&monwin,0);
  if (*(buffer + *buffer + 1) != CR)
    win_charout(CR,&monwin,0);
}

int pre_newline(channel,str)
int channel;
char *str;
{
  int len;
  
  len = get_line(str,&prewin[channel]);
  win_charout(CR,&prewin[channel],0);
  return(len);
}

void pre_charout(channel,ch)
int channel;
char ch;
{
  if (prewin[channel].column < NUM_COLUMNS-1) {
    win_charout_cntl(ch,&prewin[channel]);
  }
}

void pre_charout_nobnd(channel,ch)
int channel;
char ch;
{
  win_charout_cntl(ch,&prewin[channel]);
}

void pre_charout_cntl(channel,ch)
int channel;
char ch;
{
  if (prewin[channel].column < NUM_COLUMNS-1) {
    ch &= 0x1f;
    ch |= 0x40;
    win_attrib(att_controlchar,&prewin[channel]);
    win_charout_cntl(ch,&prewin[channel]);
    win_attrib(att_normal,&prewin[channel]);
  }
}

int cmd_newline(channel,str)
int channel;
char *str;
{
  int len;
  
  len = get_line(str,&cmdwin);
  win_charout(CR,&cmdwin,0);
  return(len);
}

void cmd_charout(channel,ch)
int channel;
char ch;
{
  if (cmdwin.column < NUM_COLUMNS-1) {
    win_charout_cntl(ch,&cmdwin,0);
  }
}

void cmd_charout_nobnd(channel,ch)
int channel;
char ch;
{
  win_charout_cntl(ch,&cmdwin);
}

int init_screen()
{
  int i;
  int j;

  infobell_flag = 0;
  cbell_flag = 1;  
  for (i = 0; i <= MONITOR; i++) {
    for (j = 0; j < 5; j++) {
      layout[i][j].win = NULL;
      layout[i][j].first_real_line = 0;
      layout[i][j].win_num_lines = 0;
    }
  }
  window_init();
  open_window(1,&statlin);
  open_window(lines_command,&cmdwin);
  layout[COMMAND][0].win = &cmdwin;
  layout[COMMAND][0].first_real_line = 0;
  layout[COMMAND][0].win_num_lines = LINES-1;
  layout[COMMAND][1].win = &statlin;
  layout[COMMAND][1].first_real_line = LINES-1;
  layout[COMMAND][1].win_num_lines = 1;
  open_window(lines_monitor,&monwin);
  layout[MONITOR][0].win = &monwin;
  layout[MONITOR][0].first_real_line = 0;
  layout[MONITOR][0].win_num_lines = LINES-1;
  layout[MONITOR][1].win = &statlin;
  layout[MONITOR][1].first_real_line = LINES-1;
  layout[MONITOR][1].win_num_lines = 1;
  for (i = 0; i < CHANNELS; i++) {
    open_window(lines_input,&prewin[i]);
    open_window(1,&statwin[i]);
    win_attrib(att_cstatline,&statwin[i]);
    open_window(lines_output,&textwin[i]);
    layout[i][0].win = &prewin[i];
    layout[i][0].first_real_line = 0;
#ifdef TNT_CHAMBER
    layout[i][0].win_num_lines = LINES/5-1;
#else
    layout[i][0].win_num_lines = LINES/3-1;
#endif
    layout[i][1].win = &statwin[i];
#ifdef TNT_CHAMBER
    layout[i][1].first_real_line = LINES/5-1;
#else
    layout[i][1].first_real_line = LINES/3-1;
#endif
    layout[i][1].win_num_lines = 1;
    layout[i][2].win = &textwin[i];
#ifdef TNT_CHAMBER
    layout[i][2].first_real_line = LINES/5;
    layout[i][2].win_num_lines = LINES-LINES/5-1;
#else
    layout[i][2].first_real_line = LINES/3;
    layout[i][2].win_num_lines = LINES-LINES/3-1;
#endif
    layout[i][3].win = &statlin;
    layout[i][3].first_real_line = LINES-1;
    layout[i][3].win_num_lines = 1;
  }
  real_screen(&layout[COMMAND][0]);
  act_mode = M_COMMAND;
  old_mode = M_COMMAND;
  act_channel = 1;
  win_attrib(att_statline,&statlin);
  for (i = 0; i < CHANNELS;i++) {
    ch_stat[i].not_disp = 0;
  }
  statlin_update();
  cmd_display(M_COMMAND,1,signon,1);
  return(0);
}

void exit_screen()
{
  int i;
  
  close_window(&cmdwin);
  close_window(&monwin);
  close_window(&statlin);
  for (i = 0; i < CHANNELS; i++) {
    close_window(&prewin[i]);
    close_window(&statwin[i]);
    close_window(&textwin[i]);
  }
  window_exit();
}
