/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993 by Mark Wahl
   For license details see documentation
   Procedures for initialization (init.c)
   created: Mark Wahl DL4YBG 94/01/11
   updated: Mark Wahl DL4YBG 94/03/20
*/

#include "tnt.h"
#include "init.h"

extern char rem_info_file[];
extern char rem_help_file[];
extern char news_file_name[];
extern char name_file_name[];
extern char tnt_dir[];
extern char remote_dir[];
extern char ctext_dir[];
extern char abin_dir[];
extern char tnt_logbookfile[];
extern int pty_timeout;
extern int auto_newline;
extern int supp_hicntl;

char tnt_initfile[80];
char tnt_logfile[80];

/* variables filled from init-file */
char device[80];
unsigned int speed;
char tnt_upfile[80];
char tnt_downfile[80];
char attc_normal;
char attc_statline;
char attc_monitor;
char attc_cstatline;
char attc_controlchar;
char attc_remote;
char attm_normal;
char attm_statline;
char attm_monitor;
char attm_cstatline;
char attm_controlchar;
char attm_remote;
int lines_command;
int lines_monitor;
int lines_input;
int lines_output;
int color;
int termcap;
char tnt_cookiefile[80];
char tnt_lockfile[80];
char tnt_ctextfile[80];
char remote_user[80];
char resy_log_file[80];

static int analyse_value(str1,str2)
char str1[];
char str2[];
{
  int tmp;
  
  if (strcmp(str1,"device") == 0) {
    strcpy(device,str2);
    return(0);
  }
  else if (strcmp(str1,"speed") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    switch (tmp) {
    case 4800:
      speed = B4800;
      return(0);
    case 9600:
      speed = B9600;
      return(0);
    case 19200:
      speed = B19200;
      return(0);
    default:
      return(1);
    }
  }
  else if (strcmp(str1,"attc_normal") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_normal = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attc_statline") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_statline = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attc_monitor") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_monitor = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attc_cstatline") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_cstatline = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attc_controlchar") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_controlchar = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attc_remote") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_remote = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_normal") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_normal = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_statline") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_statline = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_monitor") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_monitor = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_cstatline") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_cstatline = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_controlchar") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_controlchar = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_remote") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_remote = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"lines_command") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_command = tmp;
    return(0);
  }
  else if (strcmp(str1,"lines_monitor") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_monitor = tmp;
    return(0);
  }
  else if (strcmp(str1,"lines_input") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_input = tmp;
    return(0);
  }
  else if (strcmp(str1,"lines_output") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_output = tmp;
    return(0);
  }
  else if (strcmp(str1,"tnt_upfile") == 0) {
    strcpy(tnt_upfile,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_downfile") == 0) {
    strcpy(tnt_downfile,str2);
    return(0);
  }
  else if (strcmp(str1,"rem_info_file") == 0) {
    strcpy(rem_info_file,str2);
    return(0);
  }
  else if (strcmp(str1,"rem_help_file") == 0) {
    strcpy(rem_help_file,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_cookiefile") == 0) {
    strcpy(tnt_cookiefile,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_lockfile") == 0) {
    strcpy(tnt_lockfile,str2);
    return(0);
  }
  else if (strcmp(str1,"termcap") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    termcap = tmp;
    return(0);
  }
  else if (strcmp(str1,"color") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    color = tmp;
    return(0);
  }
  else if (strcmp(str1,"remote_user") == 0) {
    strcpy(remote_user,str2);
    return(0);
  }
  else if (strcmp(str1,"news_file_name") == 0) {
    strcpy(news_file_name,str2);
    return(0);
  }
  else if (strcmp(str1,"name_file_name") == 0) {
    strcpy(name_file_name,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_ctextfile") == 0) {
    strcpy(tnt_ctextfile,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_dir") == 0) {
    strcpy(tnt_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"remote_dir") == 0) {
    strcpy(remote_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"ctext_dir") == 0) {
    strcpy(ctext_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"abin_dir") == 0) {
    strcpy(abin_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_logbookfile") == 0) {
    strcpy(tnt_logbookfile,str2);
    return(0);
  }
  else if (strcmp(str1,"pty_timeout") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    pty_timeout = tmp;
    return(0);
  }
  else if (strcmp(str1,"auto_newline") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    auto_newline = tmp;
    return(0);
  }
  else if (strcmp(str1,"supp_hicntl") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    supp_hicntl = tmp;
    return(0);
  }
  else if (strcmp(str1,"resy_log_file") == 0) {
    strcpy(resy_log_file,str2);
    return(0);
  }
  else {
    return(1);
  }
}

int read_init_file(argc,argv,unlock)
int argc;
char *argv[];
int *unlock;
{
  FILE *init_file_fp;
  int file_end;
  int file_corrupt;
  char line[82];
  char str1[82];
  char str2[82];
  char tmp_str[80];
  int rslt;
  int warning;
  int wrong_usage;
  int init_file_flag;
  char *str_ptr;
  int scanned;

  strcpy(device,DEF_DEVICE);
  speed = DEF_SPEED;
  attc_normal = DEF_ATTC_NORMAL;
  attc_statline = DEF_ATTC_STATLINE;
  attc_monitor = DEF_ATTC_MONITOR;
  attc_cstatline = DEF_ATTC_CSTATLINE;
  attc_controlchar = DEF_ATTC_CONTROLCHAR;
  attc_remote = DEF_ATTC_REMOTE;
  attm_normal = DEF_ATTM_NORMAL;
  attm_statline = DEF_ATTM_STATLINE;
  attm_monitor = DEF_ATTM_MONITOR;
  attm_cstatline = DEF_ATTM_CSTATLINE;
  attm_controlchar = DEF_ATTM_CONTROLCHAR;
  attm_remote = DEF_ATTM_REMOTE;
  lines_command = DEF_LINES_COMMAND;
  lines_monitor = DEF_LINES_MONITOR;
  lines_input = DEF_LINES_INPUT;
  lines_output = DEF_LINES_OUTPUT;
  strcpy(tnt_upfile,UP_FILE);
  strcpy(tnt_downfile,DWN_FILE);
  strcpy(rem_info_file,DEF_INFO_FILE);
  strcpy(rem_help_file,DEF_HELP_FILE);
  strcpy(tnt_cookiefile,COOKIE_FILE);
  strcpy(tnt_lockfile,DEF_LOCK_FILE);
  color = DEF_COLOR;
  termcap = DEF_TERMCAP;
  strcpy(remote_user,DEF_REMOTE_USER);
  strcpy(news_file_name,DEF_NEWS_FILE_NAME);
  strcpy(name_file_name,DEF_NAME_FILE_NAME);
  strcpy(tnt_ctextfile,DEF_TNT_CTEXTFILE);
  strcpy(tnt_dir,DEF_TNT_DIR);
  strcpy(remote_dir,DEF_REMOTE_DIR);
  strcpy(ctext_dir,DEF_CTEXT_DIR);
  strcpy(abin_dir,DEF_ABIN_DIR);
  strcpy(tnt_logbookfile,DEF_TNT_LOGBOOKFILE);
  pty_timeout = DEF_PTY_TIMEOUT;
  auto_newline = DEF_AUTO_NEWLINE;
  supp_hicntl = DEF_SUPP_HICNTL;
  strcpy(resy_log_file,DEF_RESY_LOG_FILE);
  
  strcpy(tnt_initfile,INIT_FILE);
  strcpy(tnt_logfile,"");
  wrong_usage = 0;
  scanned = 1;
  while (scanned <= argc) {
    if (strcmp(argv[scanned],"-i") == 0) {
      scanned++;
      if (scanned <= argc) {
        strcpy(tnt_initfile,argv[scanned]);
      }
      else wrong_usage = 1;
    }
    else if (strcmp(argv[scanned],"-l") == 0) {
      scanned++;
      if (scanned <= argc) {
        strcpy(tnt_logfile,argv[scanned]);
      }
      else wrong_usage = 1;
    }
    else if (strcmp(argv[scanned],"-u") == 0) {
      *unlock = 1;
    }
    scanned++;
  }
  if (wrong_usage) {
    printf("Usage : tnt [-i <init-file>] [-l <log-file>]\n");
    return(1);
  }
  
  warning = 0;
  if (!(init_file_fp = fopen(tnt_initfile,"r"))) {
    str_ptr = getenv("HOME");
    if (str_ptr != NULL) {
      strcat(tmp_str,str_ptr);
      strcat(tmp_str,tnt_initfile);
      if (!(init_file_fp = fopen(tmp_str,"r"))) {
        warning = 1;
      }
    }
    else warning = 1;
  }
  if (warning) {
    printf("WARNING: %s not found\n\n",tnt_initfile);
    sleep(2);
    return(0);
  }
  file_end = 0;
  file_corrupt = 0;
  while (!file_end) {
    if (fgets(line,82,init_file_fp) == NULL) {
      file_end = 1;
    }
    else {
      if (strlen(line) == 82) {
        file_end = 1;
        file_corrupt = 1;
      }
      else {
        if (line[0] != '#') { /* ignore comment-lines */
          rslt = sscanf(line,"%s %s",str1,str2);
          switch (rslt) {
          case EOF: /* ignore blank lines */
            break;
          case 2:
            if (analyse_value(str1,str2)) {
              file_end = 1;
              file_corrupt = 1;
            }
            break;
          default:
            file_end = 1;
            file_corrupt = 1;
            break;
          }
        }
      }
    }
  }
  fclose(init_file_fp);
  if (file_corrupt) {
    printf("WARNING: tnt.ini is in wrong format\n\n");
    sleep(2);
    return(0);
  }
  else {
    return(0);
  }
}
