/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993 by Mark Wahl
   For license details see documentation
   Procedures for keyboard input (keyboard.c)
   created: Mark Wahl DL4YBG 93/07/24
   updated: Mark Wahl DL4YBG 94/03/19
*/

#include "tnt.h"
#include "keys.h"

extern int act_channel;
extern int act_mode;
extern int pause_flag[CHANNELS];
extern int tx_hiwater;
extern struct channel_stat ch_stat[CHANNELS];
extern char *term;
extern char ok_text[];
extern int resync_count;
extern int script_channel;

int chan_flag;
int tnc_command;

static int esc_flag;
static int pass_flag;
static int exit_flag;
static struct termios okbd_termios;
static struct termios nkbd_termios;
static char func_str[11];
static char input_str[80];
static char tnc_string[257];

extern void open_logfile();
extern void close_file();
extern void open_comscript();
extern void open_sendfile();
extern void break_send();
extern void rem_time();
extern void rem_cookie();
extern void cmd_shell();
extern void cmd_endshell();
extern void cmd_redir();
static void exit_tnt();
static void cmd_settime();
static void cmd_setdate();
static void cmd_flag();
static void cmd_mycall();
static void cmd_tnc();
extern void rem_ctext();
extern void rem_name();
static void cmd_chdir();
static void cmd_cwdir();
static void cmd_resync();
extern void cmd_xconnect();
#ifdef USE_SOCKET
extern void cmd_socket();
#endif

/* values for flag command */
#define F_INFOBELL 0
#define F_CBELL 1
#define F_APPEND 2
#define F_COOKIE 3
#define F_REMOTE 4
#define F_UMLAUT 5
#define F_TXECHO 6
#define F_CTEXT 8
/* variables for flag command */
int infobell_flag;
int cbell_flag;
int append_flag;
int cookie_flag;
int remote_flag;
extern int umlaut;
int txecho_flag;
int ctext_flag;

static struct com_list com_list[] = {
  {"ALFDISP" ,4,0,NULL,"A"  ,0,0},
  {"A"       ,1,0,NULL,"A"  ,0,0},
  {"DAMADIS" ,4,0,NULL,"B"  ,0,0},
  {"B"       ,1,0,NULL,"B"  ,0,0},
  {"VERSION" ,1,0,NULL,"V"  ,0,0},
  {"CHECK"   ,2,0,NULL,"@T3",0,0},
  {"@T3"     ,3,0,NULL,"@T3",0,0},
  {"CONNECT" ,1,0,NULL,"C"  ,0,0},
  {"CTEXT"   ,2,0,NULL,"U"  ,0,0},
  {"U"       ,1,0,NULL,"U"  ,0,0},
  {"DIGIPEAT",4,0,NULL,"R"  ,0,0},
  {"R"       ,1,0,NULL,"R"  ,0,0},
  {"DISCONNE",1,0,NULL,"D"  ,0,0},
  {"DAYTIME" ,5,0,NULL,"K"  ,0,0},
  {"K"       ,1,0,NULL,"K"  ,0,0},
  {"ECHO"    ,1,0,NULL,"E"  ,0,0},
  {"FLOW"    ,2,0,NULL,"Z"  ,0,0},
  {"Z"       ,1,0,NULL,"Z"  ,0,0},
  {"FRACK"   ,1,0,NULL,"F"  ,0,0},
  {"FULLDUP" ,2,0,NULL,"@D" ,0,0},
  {"@D"      ,2,0,NULL,"@D" ,0,0},
  {"MAXFRAME",3,0,NULL,"O"  ,0,0},
  {"O"       ,1,0,NULL,"O"  ,0,0},
  {"MONITOR" ,1,0,NULL,"M"  ,0,0},
  {"MYCALL"  ,2,1,cmd_mycall,"I"  ,0,0},
  {"I"       ,1,0,NULL,"I"  ,0,0},
  {"MHEARD"  ,2,0,NULL,"H"  ,0,0},
  {"H"       ,1,0,NULL,"H"  ,0,0},
  {"PERSIST" ,1,0,NULL,"P"  ,0,0},
  {"RESPTIME",3,0,NULL,"@T2",0,0},
  {"@T2"     ,3,0,NULL,"@T2",0,0},
  {"RETRY"   ,2,0,NULL,"N"  ,0,0},
  {"N"       ,1,0,NULL,"N"  ,0,0},
  {"SLOTTIME",2,0,NULL,"W"  ,0,0},
  {"W"       ,1,0,NULL,"W"  ,0,0},
  {"TXDELAY" ,1,0,NULL,"T"  ,0,0},
  {"USERS"   ,2,0,NULL,"Y"  ,0,0},
  {"Y"       ,1,0,NULL,"Y"  ,0,0},
  {"XMITOK"  ,1,0,NULL,"X"  ,0,0},
  {"BUFFERS" ,3,0,NULL,"@B" ,0,0},
  {"@B"      ,2,0,NULL,"@B" ,0,0},
  {"A1SRTT"  ,6,0,NULL,"@A1",0,0},
  {"@A1"     ,3,0,NULL,"@A1",0,0},
  {"A2SRTT"  ,6,0,NULL,"@A2",0,0},
  {"@A2"     ,3,0,NULL,"@A2",0,0},
  {"A3SRTT"  ,6,0,NULL,"@A3",0,0},
  {"@A3"     ,3,0,NULL,"@A3",0,0},
  {"IPOLL"   ,3,0,NULL,"@I" ,0,0},
  {"@I"      ,2,0,NULL,"@I" ,0,0},
  {"8BIT"    ,3,0,NULL,"@M" ,0,0},
  {"@M"      ,4,0,NULL,"@M" ,0,0},
  {"VALCALL" ,3,0,NULL,"@V" ,0,0},
  {"@V"      ,4,0,NULL,"@V" ,0,0},
  {"CHANNEL" ,4,2,NULL,"" ,0,0},
  {"S"       ,1,2,NULL,"" ,0,0},
  {"TNC"     ,3,1,cmd_tnc,"",0,0},
  {"LOGQSO"  ,4,1,open_logfile,"",RX_NORM,RX_RCV | RX_SND},
  {"LOGREC"  ,4,1,open_logfile,"",RX_NORM,RX_RCV},
  {"LOGSND"  ,4,1,open_logfile,"",RX_NORM,RX_SND},
  {"READBIN" ,5,1,open_logfile,"",RX_BIN,RX_RCV},
  {"READ"    ,4,1,open_logfile,"",RX_PLAIN,RX_RCV},
  {"READABIN",5,1,open_logfile,"",RX_ABIN,RX_RCV},
  {"CLOSE"   ,2,1,close_file,"",0,0},
  {"SENDLOG" ,5,1,open_sendfile,"",TX_NORM,0},
  {"SENDBIN" ,5,1,open_sendfile,"",TX_BIN,0},
  {"SENDABIN",5,1,open_sendfile,"",TX_ABIN,0},
  {"SEND"    ,4,1,open_sendfile,"",TX_PLAIN,0},
  {"SENDCOM" ,5,1,open_comscript,"",0,0},
  {"BREAK"   ,2,1,break_send,"",0,0},
  {"TIMESET" ,7,1,cmd_settime,"",0,0},
  {"DATESET" ,7,1,cmd_setdate,"",0,0},
  {"EXIT"    ,2,1,exit_tnt,"",0,0},
  {"INFOBELL",4,1,cmd_flag,"",F_INFOBELL,0},
  {"CBELL"   ,2,1,cmd_flag,"",F_CBELL,0},
  {"APPEND"  ,3,1,cmd_flag,"",F_APPEND,0},
  {"COOKIE"  ,4,1,cmd_flag,"",F_COOKIE,0},
  {"REMOTE"  ,3,1,cmd_flag,"",F_REMOTE,0},
  {"UMLAUT"  ,3,1,cmd_flag,"",F_UMLAUT,0},
  {"TXECHO"  ,3,1,cmd_flag,"",F_TXECHO,0},
  {"CONTEXT" ,4,1,cmd_flag,"",F_CTEXT,0},
  {"SCOOKIE" ,3,1,rem_cookie,"",0,0},
  {"SCTEXT"  ,3,1,rem_ctext,"",0,0},
  {"NAME"    ,4,1,rem_name,"",0,0},
  {"STIME"   ,3,1,rem_time,"",0,0},
  {"SHELL"   ,2,1,cmd_shell,"",0,0},
  {"ENDSHELL",4,1,cmd_endshell,"",0,0},
  {"REDIR"   ,3,1,cmd_redir,"",0,0},
  {"ENDREDIR",4,1,cmd_endshell,"",0,0},
#ifdef USE_SOCKET
  {"SOCKET"  ,3,1,cmd_socket,"",0,0},
  {"ENDSOCK" ,4,1,cmd_endshell,"",0,0},
#endif
  {"CD"      ,2,1,cmd_chdir,"",0,0},
  {"CWD"     ,3,1,cmd_cwdir,"",0,0},
  {"RESYNC"  ,4,1,cmd_resync,"",0,0},
  {"XCONNECT",2,1,cmd_xconnect,"",0,0},
  {""        ,0,-1,NULL,""   ,0,0}
};

void init_keyboard()
{
  txecho_flag = 1;
  esc_flag = 0;
  pass_flag = 0;
  chan_flag = 0;
  exit_flag = 0;
  tcgetattr(0,&okbd_termios);
  nkbd_termios = okbd_termios;
#ifdef USE_SELECT
  nkbd_termios.c_cc[VTIME] = 0;
  nkbd_termios.c_cc[VMIN] = 1;
#else
  nkbd_termios.c_cc[VTIME] = 0;
  nkbd_termios.c_cc[VMIN] = 0;
#endif
  nkbd_termios.c_cc[VSTART] = -1;
  nkbd_termios.c_cc[VSTOP] = -1;
  nkbd_termios.c_iflag = 0;
  nkbd_termios.c_iflag |= (IGNBRK|ICRNL);
  nkbd_termios.c_oflag = 0;
  nkbd_termios.c_lflag = 0;
  nkbd_termios.c_cflag |= (CS8|CREAD|CLOCAL);
#ifdef HAS_CRTSCTS
  nkbd_termios.c_cflag &= ~(CSTOPB|PARENB|PARODD|CRTSCTS|HUPCL);
#else
  nkbd_termios.c_cflag &= ~(CSTOPB|PARENB|PARODD|HUPCL);
#endif
  tcsetattr(0,TCSADRAIN,&nkbd_termios);
}

void exit_keyboard()
{
  tcsetattr(0,TCSADRAIN,&okbd_termios);
}

/* analyse input-string (command) */
void cmd_input(channel,mode,str,len,cscript)
int channel;
int mode;
char *str;
int len;
int cscript;
{
  char com_string[9];
  int found;
  int i,j;
  int invalid;
  int chan;
  
  tnc_command = 0;
  invalid = 0;
  /* delete CR at end of string and return if empty */
  if ((str[len-1] == '\n') || (str[len-1] == CR)) {
    str[len-1] = '\0';
    len--;
  }
  if (len == 0) return;
  i = 0;
  while ((i < 8) && (i < len) && (*str != SPACE)) {
    com_string[i] = *str++;
    if ((com_string[i] > 0x60) && (com_string[i] < 0x7E))
      com_string[i] &= 0xDF;
    i++;
  }
  com_string[i]= '\0';
  if ((i != len) && (*str != SPACE)) {
    invalid = 1;
  }
  len -= i;
  if (!invalid) {
    /* find string in command-list */
    found = 0;
    j = 0;
    while ((com_list[j].ext_com != -1) && !found) {
      if (strstr(com_list[j].string,com_string) == com_list[j].string) {
        /* string found, now check if length valid */
        if (strlen(com_string) >= com_list[j].len) {
          found = 1;
        }
      }
      if (!found) j++;
    }
    if (found) {
      /* delete spaces before parameters */
      while (*str == SPACE) {
        str++;
        len--;
      }
      switch (com_list[j].ext_com) {
      case 0: /* tnc command */
        tnc_command = 1;
        strcpy(tnc_string,com_list[j].new_string);
        if (len)
          strcat(tnc_string,str);
        len = strlen(tnc_string);
        queue_cmd_data(channel,X_COMM,len,mode,tnc_string);
        break;
      case 1: /* external command */
        (*com_list[j].func)(com_list[j].par1,com_list[j].par2,
                            channel,len,mode,str);
        break;
      case 2: /* change of channel */
        chan = atoi(str);
        if ((chan >= 0) && (chan < CHANNELS)) {
          if (cscript) { /* change command script channel */
            script_channel = chan;
          }
          else { /* change channel */
            sel_connect(chan);
            statlin_update();
          }
        }
        break;
      }
      return;
    }
  }
  cmd_display(mode,channel,"Invalid command",1);
  return;
}

void exit_tnt(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  cmd_display(mode,channel,"Do you really want to exit (y/n)? ",0);
  exit_flag = 1;
}

/* analyse input-string (data) */
void dat_input(channel,str,len)
int channel;
char *str;
int len;
{
  int flag;
  
  flag = 0;
  queue_cmd_data(channel,X_DATA,len,flag,str);
  write_file(channel,len,RX_SND,str);
}

void keyboard_server(state,ch)
int *state;
char *ch;
{
  int i;
  int found;
  int analysed;
  int input_len;
  char temp_str[100];
  int num_func;
  struct func_keys *spec_keys;
  

  if (exit_flag) {
    switch (*ch) {
      case 'Y':
      case 'y':
        *state = S_TNT_DWN;
        open_dwnfile(state);
        exit_flag = 0;
        cmd_display(act_mode,act_channel,"Y",1);
        break;
      case 'N':
      case 'n':
        exit_flag = 0;
        cmd_display(act_mode,act_channel,"N",1);
        break;
    }
    return;
  }
  else if (chan_flag) {
    if ((*ch >= '0') && (*ch <= '9')) {
      chan_flag = 0;
      sel_connect((int)(*ch-'0'));
      statlin_update();
      return;
    }
    else if (*ch == '@') {
      chan_flag = 0;
      sel_connect((int)(*ch-'@'+10));
      statlin_update();
      return;
    }
    else return; /* wait for valid channel number */
  }
  if (esc_flag) {
    found = 0;
    if (strcmp(term,"xterm") == 0) {
      spec_keys = special_keys_xterm;
      num_func = NUMFUNC_XTERM;
    }
    else {
      spec_keys = special_keys;
      num_func = NUMFUNC;
    }
    func_str[esc_flag] = *ch;
    func_str[esc_flag + 1] = '\0';
    for (i = 0; (i < num_func) && (!found); i++) {
      if (strstr(spec_keys[i].code,func_str) == spec_keys[i].code) {
        found = 1;
        if (strlen(spec_keys[i].code) == (esc_flag + 1)) {
          esc_flag = 0;
          if (spec_keys[i].sel_channel) {
            sel_connect((int)spec_keys[i].res_code);
            return;
          }
          else {
            *ch = spec_keys[i].res_code;
          }
        }
        else {
          if (++esc_flag > 10) esc_flag = 0;
          return;
        }
      }
    }
    if (!found) {
      esc_flag = 0;
      return;
    }
  }
  if (*ch < SPACE) {
    analysed = 1;
    switch (*ch) {
      case ESC:
        func_str[0] = ESC;
        esc_flag = 1;
        break;
      case C_MONITOR:
        sel_monitor();
        break;
      case C_COMMAND:
        sel_command();
        break;
      case C_CONNECT:
        sel_connect(act_channel);
        break;
      case C_CHANNEL:
        chan_flag = 1;
        statlin_update();
        break;
      default:
        analysed = 0;
        break;
    }
    if (analysed) return;
  }
  if (*state == S_NORMAL) {
    switch (act_mode) {
    case M_CONNECT:
      if (pass_flag) {
        pre_charout_cntl(act_channel,*ch);
        pass_flag = 0;
      }
      else if ((*ch < SPACE) || (*ch == ESC2) || (*ch == DEL)) {
        switch (*ch) {
          case LF:
            if ((tx_hiwater) || (!data_allowed(act_channel))) {
              beep();
            }
            else {
              input_len = pre_newline(act_channel,input_str);
              if ((input_len > 1) && (input_str[0] == ':'))
                cmd_input(act_channel,act_mode,input_str + 1,input_len - 1,0);
              else {
                dat_input(act_channel,input_str,input_len);
                if (txecho_flag) {
                  insert_cr_rx(act_channel);
                  temp_str[0] = (char)((input_len - 1) & 0x00ff);
                  strncpy(temp_str+1,input_str,input_len);
                  rem_data_display_len(act_channel,temp_str);
                }
              }
            }
            break;
          case C_PASS:
            pass_flag = 1;
            break;
          case C_PAUSE:
            pause_flag[act_channel] ^= 1;
            statlin_update();
            break;
          case C_WINUP:
            real_window_up(2,10);
            break;
          case C_WINDWN:
            real_window_down(2,10);
            break;
          case C_LINEUP:
            real_window_up(2,1);
            break;
          case C_LINEDWN:
            real_window_down(2,1);
            break;
          case C_CULEFT:
          case C_CUUP:
          case C_CUDWN:
          case C_STLINE:
          case C_EOLINE:
          case C_INSERT:
          case C_DELLINE:
          case C_DELCHAR:
          case BS:
          case DEL:
            pre_charout_nobnd(act_channel,*ch);
            break;
          case C_CURIGHT:
            pre_charout(act_channel,*ch);
            break;
          default:
            pre_charout_cntl(act_channel,*ch);
            break;
        }
      }
      else {
        pre_charout(act_channel,*ch);
      }
      break;
    case M_COMMAND:
      if ((*ch < SPACE) || (*ch == ESC2) || (*ch == DEL)) {
        switch (*ch) {
          case LF:
            if (tx_hiwater)
              beep();
            else {  
              input_len = cmd_newline(act_channel,input_str);
              cmd_input(act_channel,act_mode,input_str,input_len,0);
            }
            break;
          case C_CULEFT:
          case C_CUUP:
          case C_CUDWN:
          case C_STLINE:
          case C_EOLINE:
          case C_INSERT:
          case C_DELLINE:
          case C_DELCHAR:
          case BS:
          case DEL:
            cmd_charout_nobnd(act_channel,*ch);
            break;
          case C_CURIGHT:
            cmd_charout(act_channel,*ch);
            break;
        }
      }
      else {
        cmd_charout(act_channel,*ch);
      }
      break;
    case M_MONITOR:
      switch (*ch) {
        case C_PAUSE:
          pause_flag[0] ^= 1;
          statlin_update();
          break;
        case C_WINUP:
          real_window_up(0,20);
          break;
        case C_WINDWN:
          real_window_down(0,20);
          break;
        case C_LINEUP:
          real_window_up(0,1);
          break;
        case C_LINEDWN:
          real_window_down(0,1);
          break;
      }
      break;
    }
  }
}

static void cmd_settime(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  struct tm timestr;
  time_t timeval;
  char tmpstr[10];
    
  timeval = time(&timeval);
  timestr = *localtime(&timeval);
  sprintf(tmpstr,"K %2.2u:%2.2u:%2.2u",
          timestr.tm_hour,timestr.tm_min,timestr.tm_sec);
  queue_cmd_data(channel,X_COMM,strlen(tmpstr),mode,tmpstr);
  cmd_display(mode,channel,tmpstr,1);
  tnc_command = 1;
}

static void cmd_setdate(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  struct tm timestr;
  time_t timeval;
  char tmpstr[10];
    
  timeval = time(&timeval);
  timestr = *localtime(&timeval);
  sprintf(tmpstr,"K %2.2u.%2.2u.%2.2u",
          timestr.tm_mday,timestr.tm_mon+1,timestr.tm_year);
  queue_cmd_data(channel,X_COMM,strlen(tmpstr),mode,tmpstr);
  cmd_display(mode,channel,tmpstr,1);
  tnc_command = 1;
}

static void cmd_flag(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int *flag;
  int i;
  char ch;
  
  switch (par1) {
  case F_INFOBELL:
    flag = &infobell_flag;
    break;
  case F_CBELL:
    flag = &cbell_flag;
    break;
  case F_APPEND:
    flag = &append_flag;
    break;
  case F_COOKIE:
    flag = &cookie_flag;
    break;
  case F_REMOTE:
    flag = &remote_flag;
    break;
  case F_UMLAUT:
    flag = &umlaut;
    break;
  case F_TXECHO:
    flag = &txecho_flag;
    break;
  case F_CTEXT:
    flag = &ctext_flag;
    break;
  }
  if (len == 0) {
    if (*flag == 0)
      cmd_display(mode,channel,"OFF",1);
    else
      cmd_display(mode,channel,"ON",1);
    return;
  }
  for (i = 0; i < len; i++) {
    ch = str[i];
    if ((ch > 0x40) && (ch < 0x5b)) {
      ch |= 0x20;
      str[i] = ch;
    }
  }    
  if (len >= 2) {
    if (strncmp(str,"on",2) == 0) {
      *flag = 1;
      cmd_display(mode,channel,"Ok",1);
      return;
    }
  } 
  if (len >= 3) {
    if (strncmp(str,"off",3) == 0) {
      *flag = 0;
      cmd_display(mode,channel,"Ok",1);
      return;
    }
  }
  cmd_display(mode,channel,"INVALID VALUE",1);
}

/* Set permanent mycall */
static void cmd_mycall(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int i;
  
  strcpy(tnc_string,"I");
  if (len) {
    if (strcmp(str,"$") == 0) {
      ch_stat[channel].mycall[0] = '\0';
    }
    else {
      tnc_command = 1;
      for (i=0;i<9;i++) {
        ch_stat[channel].mycall[i] = toupper(str[i]);
      }
      strcat(tnc_string,str);
      len = strlen(tnc_string);
      queue_cmd_data(channel,X_COMM,len,mode,tnc_string);
    }
  }
  else {
    if (ch_stat[channel].mycall[0] == '\0') 
      cmd_display(mode,channel,"<empty>",1);
    else
      cmd_display(mode,channel,ch_stat[channel].mycall,1);
  }
}

/* Send TNC command */
static void cmd_tnc(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int i;
  
  if (len) {
    tnc_command = 1;
    strcpy(tnc_string,str);
    len = strlen(tnc_string);
    queue_cmd_data(channel,X_COMM,len,mode,tnc_string);
  }
  else {
    cmd_display(mode,channel,"INVALID TNC STRING",1);
  }
}

/* Change directory */
static void cmd_chdir(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int result;
  char *dir;

  dir = str;
  if (strlen(dir) == 0) {
    dir = getenv("HOME");
  }  
  if ((result = chdir(dir)) == -1) {
    cmd_display(mode,channel,"Can't change directory",1);
  }
  else {
    cmd_display(mode,channel,ok_text,1);
  }
}

/* Display working directory */
static void cmd_cwdir(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  char dir[80];
  
  if (getcwd(dir,80) == NULL) {
    cmd_display(mode,channel,"Can't get directory name",1);
  }
  else {
    cmd_display(mode,channel,dir,1);
  }
}

/* Display number of resyncs up to now */
static void cmd_resync(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  char number[20];
  
  sprintf(number,"%d",resync_count);
  cmd_display(mode,channel,number,1);
}
