/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993 by Mark Wahl
   For license details see documentation
   Main procedure (main.c)
   created: Mark Wahl DL4YBG 93/08/07
   updated: Mark Wahl DL4YBG 94/03/21
*/
#define TNT_DATE "94/03/21"

#include "tnt.h"

extern int serial;
extern char device[];
extern unsigned int speed;
extern char tnt_logfile[];

int act_channel;
int act_mode;
int act_state;

char signon[] =
"TNT V0.7e Packet Radio Hostmode Terminal\015"
#ifdef TNT_LINUX
"LINUX"
#endif
#ifdef TNT_ISC
"ISC"
#endif
"-Version, Created: "TNT_DATE"\015"
"Copyright (C) 1993 by Mark Wahl, DL4YBG\015"
"For license details see documentation.\015";

char rem_ver_str[] =
"TNT/"
#ifdef TNT_LINUX
"LINUX"
#endif
#ifdef TNT_ISC
"ISC"
#endif
" V0.7e ("TNT_DATE", DL4YBG)";


int 
main(argc,argv)
int argc;
char *argv[];
{
  int len;
  char rdbuffer[259];
  char ch;
  int unlock;
#ifdef USE_SELECT
  struct fd_set fdmask;
  struct timeval timevalue;
  int served;
  int max_fd;
  int count;
#else
  int kbd_res;
  int ser_res;
  int ser_resp;
  int shell_res;
#ifndef HAS_USLEEP
  struct timeval timevalue;
#endif
#endif

  umask(0); /* don't filter file-permissions */
  if (read_init_file(argc,argv,&unlock))
    exit(1);
  if (term_init())
    exit(1);
  
  if (init_serial(device,speed,unlock))
    exit(1);

  signal(SIGHUP, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);
  signal(SIGTTIN, SIG_IGN);
  signal(SIGTTOU, SIG_IGN);

  init_file();
  init_screen();  
  init_keyboard();
  init_conv();
  init_remote();
  init_cookie();
  init_crc();
  init_shell();
  init_log();
  
  if (strcmp(tnt_logfile,"") != 0) {
    open_logfile(RX_NORM,RX_RCV,1,strlen(tnt_logfile),
                 M_CMDSCRIPT,tnt_logfile);
  }
  
#ifdef USE_SELECT

  act_state = S_INIT;
  len = 1;
  serial_server(&act_state,rdbuffer,len);
  
  while (act_state) {
    FD_ZERO(&fdmask);
    FD_SET(0,&fdmask);
    FD_SET(serial,&fdmask);
    max_fd = serial + 1;
    shell_fdset(&max_fd,&fdmask);
    timevalue.tv_usec = 0;
    timevalue.tv_sec = 2;
    count = select(max_fd,&fdmask,
                  (struct fd_set *) 0,(struct fd_set *) 0,&timevalue);
    if (count == -1) {
      continue;
    }
    served = 0;
    if (FD_ISSET(0,&fdmask)) {
      if (read(0,&ch,1) == 1) {
        keyboard_server(&act_state,&ch);
      }
      served = 1;
    }
    if (FD_ISSET(serial,&fdmask)) {
      if (len = read(serial,rdbuffer,259)) {
        serial_server(&act_state,rdbuffer,len);
      }
      served = 1;
    }
    if (!shell_receive(&fdmask)) {
      served = 1;
    } 
    if (!served) {
      resync_tnc(&act_state);
    }
  }

#else

  act_state = S_INIT;
  ser_resp = 0;
  len = 1;
  serial_server(&act_state,rdbuffer,len);

  while (act_state) {
    if (ser_resp > MAXPOLL) {
      resync_tnc(&act_state);
      ser_resp = 0;
    }
    ser_res = 1;
    if (len = read(serial,rdbuffer,259)) {
      serial_server(&act_state,rdbuffer,len);
      ser_res = 0;
      ser_resp = 0;
    }
    kbd_res = 1;
    if (read(0,&ch,1)) {
      keyboard_server(&act_state,&ch);
      kbd_res = 0;
    }
    shell_res = shell_receive();
    /* if no serial processing and no key pressed: sleep */
    if (ser_res && kbd_res && shell_res) {
#ifdef HAS_USLEEP
      usleep(1000); /* sleep for 1ms */
#else
      timevalue.tv_usec = 1000; /* sleep for 1ms */
      timevalue.tv_sec = 0;
      select(0,(struct fd_set *) 0,
               (struct fd_set *) 0,
               (struct fd_set *) 0,
               &timevalue);
#endif
      ser_resp++;
    }
  }
  
#endif  

  exit_log();
  exit_shell();
  exit_file();
  exit_keyboard();
  exit_screen();
  if (exit_serial())
    exit(1);

  exit(0);
}
