/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993 by Mark Wahl
   For license details see documentation
   Procedures for shell-login (shell.c)
   created: Mark Wahl DL4YBG 94/01/22
   updated: Mark Wahl DL4YBG 94/03/20
*/

#include "tnt.h"
#include "shell.h"
#include <pwd.h>
#include <utmp.h>

#ifdef USE_SOCKET
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

#ifdef TNT_CHAMBER
#define BINTRX "/tmp"
#ifndef _NFILE
#define _NFILE FD_SETSIZE
#endif
#endif

#define BUFFER_LEN 256

static struct shell_stat sh_stat[CHANNELS];
extern struct channel_stat ch_stat[CHANNELS];
int pty_timeout;

#ifdef USE_SOCKET
static struct sockaddr_in cli_addr;
static struct sockaddr_in serv_addr;
static int clilen;
#endif

#ifdef USE_SOCKET
static char sh_active_txt[] = "Shell, redirection or socket already active";
#else
static char sh_active_txt[] = "Shell or redirection already active";
#endif

static long secclock()
{
  struct timeval tv;
  struct timezone tz;
  
  gettimeofday(&tv,&tz);
  return(tv.tv_sec);
}

static int find_pty(numptr, slave)
int *numptr;
char *slave;
{

  char master[80];
  int fd;
  int num;

  for (num = 0; num < NUMPTY; num++) {
    pty_name(master, MASTERPREFIX, num);
#ifdef USE_SELECT
    if ((fd = open(master, O_RDWR, 0600)) >= 0) {
#else
    if ((fd = open(master, O_RDWR | O_NONBLOCK, 0600)) >= 0) {
#endif
      *numptr = num;
      pty_name(slave, SLAVEPREFIX, num);
      return fd;
    }
  }
  return (-1);
}

static void restore_pty(id)
const char *id;
{
  char filename[80];

  sprintf(filename, "%s%s", MASTERPREFIX, id);
  chown(filename, 0, 0);
  chmod(filename, 0666);
  sprintf(filename, "%s%s", SLAVEPREFIX, id);
  chown(filename, 0, 0);
  chmod(filename, 0666);
}

void fixutmpfile()
{

#ifdef USER_PROCESS

  struct utmp *up;

  while (up = getutent())
    if (up->ut_type == USER_PROCESS && kill(up->ut_pid, 0)) {
      restore_pty(up->ut_id);
      up->ut_user[0] = 0;
      up->ut_type = DEAD_PROCESS;
#ifdef HAS_UTEXIT
      up->ut_exit.e_termination = 0;
      up->ut_exit.e_exit = 0;
#endif
      up->ut_time = secclock();
      pututline(up);
    }
  endutent();

#endif
}

static int callvalid(call)
const char *call;
{
  int d, l;

  l = strlen(call);
  if (l < 3 || l > 6) return 0;
  if (isdigit(uchar(call[0])) && isdigit(uchar(call[1]))) return 0;
  if (!(isdigit(uchar(call[1])) || isdigit(uchar(call[2])))) return 0;
  if (!isalpha(uchar(call[l-1]))) return 0;
  d = 0;
  for (; *call; call++) {
    if (!isalnum(uchar(*call))) return 0;
    if (isdigit(uchar(*call))) d++;
  }
  if (d < 1 || d > 2) return 0;
  return 1;
}

static char *find_user_name(name)
const char *name;
{

  char *cp;
  static char username[128];

  for (; ; ) {
    while (*name && !isalnum(uchar(*name))) name++;
    for (cp = username; isalnum(uchar(*name)); *cp++ = tolower(uchar(*name++))) ;
    *cp = 0;
    if (!*username) return DEFAULTUSER;
    if (callvalid(username)) return username;
  }
}

struct passwd *getpasswdentry(name, create)
const char *name;
int create;
{
  struct passwd *pw;

  /* Search existing passwd entry */

  if (pw = getpwnam(name)) return pw;
  else return 0;
}

#ifdef TNT_CHAMBER
void disc_frontend(channel)
int channel;
{
  int flag;
  char ans_str[10];
  int ans_len;
  
  flag = 0;
  strcpy(ans_str,"D");
  ans_len = strlen(ans_str);
  queue_cmd_data(channel,X_COMM,ans_len,M_CMDSCRIPT,ans_str);
}
#endif

/* open a shell on the current channel */
void cmd_shell(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{

  char *env = 0;
  char slave[80];
  int i;
  struct passwd *pw;
  struct termios termios;
  struct utmp utmpbuf;

  if (sh_stat[channel].active) {
    cmd_display(mode,channel,sh_active_txt,1);
    return;
  }
  if ((sh_stat[channel].pty = find_pty(&sh_stat[channel].num, slave)) < 0) {
    cmd_display(mode,channel,"Can't open pseudo terminal",1);
    return;
  }
  sh_stat[channel].active = A_SHELL;
  sh_stat[channel].mode = mode;
  sh_stat[channel].buflen = 0;
  for (i=0;i < len;i++) {
    str[i] = tolower(str[i]);
  }
  sh_stat[channel].lfcrconv = !(strcmp(str,"trans") == 0);
  strcpy(sh_stat[channel].id, slave + strlen(slave) - 2);
  if (!(sh_stat[channel].pid = fork())) {
    pw = getpasswdentry(find_user_name(ch_stat[channel].call), 1);
    if (!pw || pw->pw_passwd[0]) pw = getpasswdentry(DEFAULTUSER, 0);
    for (i = 0; i < FD_SETSIZE; i++) close(i);
    setsid();
    open(slave, O_RDWR, 0666);
    dup(0);
    dup(0);
    chmod(slave, 0622);
    if (sh_stat[channel].lfcrconv) {
      memset((char *) &termios, 0, sizeof(termios));
      termios.c_iflag = ICRNL | IXOFF;
      termios.c_oflag = OPOST | TAB3 | ONLRET;
      termios.c_oflag &= ~(OCRNL | ONLCR | ONOCR);
      termios.c_cflag = CS8 | CREAD | CLOCAL;
      termios.c_lflag = ISIG | ICANON;
      termios.c_cc[VINTR]  = 127;
      termios.c_cc[VQUIT]  =  28;
      termios.c_cc[VERASE] =   8;
      termios.c_cc[VKILL]  =  24;
      termios.c_cc[VEOF]   =   4;
      cfsetispeed(&termios, B1200);
      cfsetospeed(&termios, B1200);
    }
    else {
      tcgetattr(0,&termios);
    }
    tcsetattr(0, TCSANOW, &termios);
    if (!pw || pw->pw_passwd[0]) exit(1);
#ifdef LOGIN_PROCESS
    memset(&utmpbuf, 0, sizeof(utmpbuf));
    strcpy(utmpbuf.ut_user, "LOGIN");
    strcpy(utmpbuf.ut_id, sh_stat[channel].id);
    strcpy(utmpbuf.ut_line, slave + 5);
    utmpbuf.ut_pid = getpid();
    utmpbuf.ut_type = LOGIN_PROCESS;
    utmpbuf.ut_time = secclock();
#ifdef HAS_UTHOST 
    strncpy(utmpbuf.ut_host,"AX25", sizeof(utmpbuf.ut_host));
#endif
    pututline(&utmpbuf);
    endutent();
#endif
#ifdef TNT_CHAMBER
    execl("/bin/login.pr", "login.pr", (char *) 0);
#else
    execle("/bin/login", "login", pw->pw_name, (char *) 0, &env);
#endif
    exit(1);
  }
  cmd_display(mode,channel,"Ok",1);
  return;
}

/* redirect the current channel */
void cmd_redir(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{

  char ans_str[80];

  if (sh_stat[channel].active) {
    cmd_display(mode,channel,sh_active_txt,1);
    return;
  }
#ifdef USE_SELECT
  if ((sh_stat[channel].pty = open(str,O_RDWR,0600)) < 0) {
#else
  if ((sh_stat[channel].pty = open(str,O_RDWR|O_NONBLOCK,0600)) < 0) {
#endif
    sprintf(ans_str,"Can't redirect to %s",str);
    cmd_display(mode,channel,ans_str,1);
    return;
  }
  sh_stat[channel].active = A_REDIR;
  sh_stat[channel].mode = mode;
  sh_stat[channel].pid = 0;
  sh_stat[channel].lfcrconv = 0;
  sh_stat[channel].buflen = 0;
  sprintf(ans_str,"Redirect to %s",str);
  cmd_display(mode,channel,ans_str,1);
}

#ifdef USE_SOCKET
/* redirect the current channel to a socket */
void cmd_socket(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{

  char ans_str[80];
  int port;
  
  if (sscanf(str,"%d",&port) != 1) {
    cmd_display(mode,channel,"Invalid port",1);
    return;
  }
  if (sh_stat[channel].active) {
    cmd_display(mode,channel,sh_active_txt,1);
    return;
  }
  if ((sh_stat[channel].sockfd = socket(AF_INET,SOCK_STREAM,0)) < 0) {
    cmd_display(mode,channel,"Can't open stream socket",1);
    return;
  }
#ifndef USE_SELECT  
  fcntl(sh_stat[channel].sockfd,F_SETFL,O_NONBLOCK);
#endif
  bzero((char *) &serv_addr,sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(port);
  if (bind(sh_stat[channel].sockfd, (struct sockaddr *) &serv_addr,
           sizeof (serv_addr)) < 0) {
    sprintf(ans_str,"Can't bind port %d",port);
    cmd_display(mode,channel,ans_str,1);
    close(sh_stat[channel].sockfd);
    return;
  }
  listen(sh_stat[channel].sockfd,5);

  sh_stat[channel].active = A_SOCKET;
  sh_stat[channel].mode = mode;
  sh_stat[channel].pid = 0;
  sh_stat[channel].lfcrconv = 1;
  sh_stat[channel].buflen = 0;
  sh_stat[channel].connected = 0;
  sprintf(ans_str,"Listening at port %d",port);
  cmd_display(mode,channel,ans_str,1);
}
#endif

void pty_send_data(channel,buffer,len)
int channel;
char *buffer;
int len;
{
  int tmp_len;
  
  if (len > PACKETSIZE) return; /* sanity check */
  if ((sh_stat[channel].buflen == 0) && (len == PACKETSIZE)) {
    dat_input(channel,buffer,len);
  }
  else if ((tmp_len = sh_stat[channel].buflen + len) > PACKETSIZE) {
    if (sh_stat[channel].buflen) {
      dat_input(channel,sh_stat[channel].buffer,sh_stat[channel].buflen);
    }
    memcpy(sh_stat[channel].buffer,buffer,len);
    sh_stat[channel].buflen = len;
    sh_stat[channel].bufupdate = time(NULL);
  }
  else {
    memcpy((sh_stat[channel].buffer + sh_stat[channel].buflen),
           buffer,len);
    sh_stat[channel].buflen = tmp_len;       
    sh_stat[channel].bufupdate = time(NULL);
  }
}

void pty_check_timeout(channel)
int channel;
{
  if (!sh_stat[channel].buflen) return;
  if ((time(NULL) - sh_stat[channel].bufupdate) >= pty_timeout) {
    dat_input(channel,sh_stat[channel].buffer,sh_stat[channel].buflen);
    sh_stat[channel].buflen = 0;
  }
}

static int read_data_pty(channel,buffer)
int channel;
char buffer[];
{
  int len;
#ifdef READ_NONBLOCK
  int fc;
#endif
#ifdef TNT_CHAMBER
  char buffer2[BUFFER_LEN+2];
  char tmp[100];
  static long xx;
  FILE *fp;
#endif

#ifdef USE_SOCKET
  if ((sh_stat[channel].active == A_SOCKET) &&
      (!sh_stat[channel].connected)) return(0);
#endif
#ifdef READ_NONBLOCK
  fcntl(sh_stat[channel].pty,F_SETFL,O_NONBLOCK);
#endif
  len = read(sh_stat[channel].pty,buffer,BUFFER_LEN);
#ifdef READ_NONBLOCK
  fc=fcntl(sh_stat[channel].pty,F_GETFL,NULL);
  fcntl(sh_stat[channel].pty,F_SETFL,fc & ~O_NONBLOCK);
#endif
  if (len > 0) {
#ifdef TNT_CHAMBER
    if (sh_stat[channel].active == A_FRONTEND) {
      sprintf(tmp,"%s/.bin.%d.%d",BINTRX,getpid(),channel);
      if(access(tmp,0)!=0) {
        len=conv_lftonil(buffer,len);
        xx=0L;
        buffer2[0] = len-1;
        memcpy(&buffer2[1],buffer,len);
        rem_data_display_len(channel,buffer2);
      }
      else {
        xx+=(long)len;
        sprintf(buffer2,"RBIN-Frontend: %ld",xx);
        rem_stat_display(channel,buffer2);
      }
    }
    else {
      if (sh_stat[channel].lfcrconv) {
        conv_lftocr(buffer,len);
      }
    }
#else
    if (sh_stat[channel].lfcrconv) {
      conv_lftocr(buffer,len);
    }
#endif
    pty_send_data(channel,buffer,len);
#ifdef TNT_CHAMBER
    if((sh_stat[channel].active==A_FRONTEND) && (access(tmp,0)!=0)) {
      fp=fopen("/pr/log/FRONTEND.log","a");
      if(fp!=0)  {
        strncpy(buffer2,buffer,len);
        buffer2[strlen(buffer2)]=0;
        conv_crtolf(buffer2,len);
        fprintf(fp,"%s",buffer2);
        fclose(fp);
      }
    }
#endif
    return(1);
  }
  else {
#ifdef USE_SOCKET  
    if ((len == 0) && (sh_stat[channel].active == A_SOCKET)){
      if (sh_stat[channel].connected) {
        close(sh_stat[channel].pty);
        sh_stat[channel].connected = 0;
      }
    }
#endif
    return(0);
  }
}

void close_shell(channel,report)
int channel;
int report;
{

#ifndef UTMP_FILE
#define UTMP_FILE       "/etc/utmp"
#endif

#ifndef WTMP_FILE
#define WTMP_FILE       "/var/adm/wtmp"
#endif

  char slave[80];
  int fdut = -1;
  struct utmp utmpbuf;
  char ans_str[80];
  int result;
  char buffer[BUFFER_LEN];
#ifndef USE_SELECT
  int nodata;
#endif

  if (!sh_stat[channel].active) return;
#ifdef USE_SOCKET
  if (sh_stat[channel].active == A_SOCKET) return;
#endif
  /* read all remaining data from pty */
#ifdef USE_SELECT
  fcntl(sh_stat[channel].pty,F_SETFL,O_NONBLOCK);
#endif
  while (read_data_pty(channel,buffer) == 1);
  if (sh_stat[channel].buflen) {
    dat_input(channel,sh_stat[channel].buffer,sh_stat[channel].buflen);
  }
  if (sh_stat[channel].pty > 0) {
    ioctl(sh_stat[channel].pty, TCFLSH, 2);
    close(sh_stat[channel].pty);
#ifdef TNT_CHAMBER
    if ((sh_stat[channel].active == A_SHELL) ||
       (sh_stat[channel].active == A_FRONTEND))
#else
    if (sh_stat[channel].active == A_SHELL)
#endif
      restore_pty(sh_stat[channel].id);
  }
#ifdef TNT_CHAMBER
  if (((sh_stat[channel].active == A_SHELL) ||
       (sh_stat[channel].active == A_FRONTEND)) &&
      (sh_stat[channel].pid > 0)) {
#else
  if ((sh_stat[channel].active == A_SHELL) && (sh_stat[channel].pid > 0)) {
#endif
    kill(-sh_stat[channel].pid, SIGHUP);
    pty_name(slave, SLAVEPREFIX, sh_stat[channel].num);
#ifdef TNT_CHAMBER
    if (sh_stat[channel].active == A_FRONTEND) {
      char tmp[100];
      sprintf(tmp,"%s/.bin.%d.%d",BINTRX,getpid(),channel);
      if(access(tmp,0)==0) unlink(tmp);
      disc_frontend(channel);
    }
#endif
    if ((fdut = open(UTMP_FILE, O_RDWR, 0644)) >= 0) {
      while (read(fdut, &utmpbuf, sizeof(utmpbuf)) == sizeof(utmpbuf)) {
	if (!strncmp(utmpbuf.ut_line, slave + 5, sizeof(utmpbuf.ut_line))) {
	  utmpbuf.ut_name[0] = 0;
#ifdef HAS_UTHOST	  
	  utmpbuf.ut_host[0] = 0;
#endif
	  utmpbuf.ut_time = secclock();
#ifdef DEAD_PROCESS
	  utmpbuf.ut_type = DEAD_PROCESS;
#ifdef HAS_UTEXIT
	  utmpbuf.ut_exit.e_termination = 0;
	  utmpbuf.ut_exit.e_exit = 0;
#endif
#endif
	  lseek(fdut, -sizeof(utmpbuf), SEEK_CUR);
	  write(fdut, &utmpbuf, sizeof(utmpbuf));
	  close(fdut);
	  if ((fdut = open(WTMP_FILE, O_WRONLY | O_CREAT | O_APPEND, 0644)) >= 0)
	    write(fdut, &utmpbuf, sizeof(utmpbuf));
	  break;
	}
      }	
    }
    if (fdut >= 0) close(fdut);
  }
  if (report) {
    switch (sh_stat[channel].active) {
    case A_SHELL:
      strcpy(ans_str,"Shell closed");
      break;
    case A_REDIR:
      strcpy(ans_str,"Redirection closed");
      break;
#ifdef TNT_CHAMBER
    case A_FRONTEND:
      strcpy(ans_str,"Frontend closed");
      break;
#endif
    }
    cmd_display(sh_stat[channel].mode,channel,ans_str,1);
  }
  sh_stat[channel].active = 0;
}

#ifdef USE_SOCKET
static void close_socket(channel,report)
int channel;
int report;
{
  char ans_str[80];

  if (!sh_stat[channel].active) return;
  if (sh_stat[channel].active != A_SOCKET) return;
  if (sh_stat[channel].connected) {
    close(sh_stat[channel].pty);
  }
  close(sh_stat[channel].sockfd);
  if (report) {
    strcpy(ans_str,"Socket closed");
    cmd_display(sh_stat[channel].mode,channel,ans_str,1);
  }
  sh_stat[channel].active = 0;
}
#endif

/* close a shell or redir on the current channel */
void cmd_endshell(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
#ifdef USE_SOCKET
  if (sh_stat[channel].active == A_SOCKET) {
    close_socket(channel,1);
  }
#endif
  close_shell(channel,1);
}

int write_pty(channel,len,str)
int channel;
int len;
char *str;
{
#ifdef TNT_CHAMBER
  char buffer[300];
  FILE *fp;
  int i;
#endif

  if (!sh_stat[channel].active) return(0);
  if (sh_stat[channel].lfcrconv) {
    conv_crtolf(str,len);
  }
  if (write(sh_stat[channel].pty,str,len) < len) {
    close_shell(channel,1);
  }
#ifdef TNT_CHAMBER
  if (sh_stat[channel].active==A_FRONTEND) {
    buffer[0] = len-1;
    memcpy(&buffer[1],str,len);
    rem_data_display_len(channel,buffer);
    
    fp=fopen("/pr/log/FRONTEND.log","a");
    if (fp!=0) {
      for(i=0;i<len;i++) {
      if(isprint(str[i]))
         fprintf(fp,"%c",str[i]);
      }
      fprintf(fp,"\n");
      fclose(fp);
    }
  }
#endif
  return(1);
}

#ifdef USE_SELECT
void shell_fdset(max_fd,fdmask)
int *max_fd;
struct fd_set *fdmask;
{
  int channel;
  int fd;
  
  for (channel=0;channel<CHANNELS;channel++) {
    if ((sh_stat[channel].active) && senddata_allowed(channel)) {
#ifdef USE_SOCKET
      if ((sh_stat[channel].active == A_SOCKET) &&
          (!sh_stat[channel].connected)) {
        fd = sh_stat[channel].sockfd;
      }
      else {
        fd = sh_stat[channel].pty;
      }
#else
      fd = sh_stat[channel].pty;
#endif
      FD_SET(fd,fdmask);
      if (fd > ((*max_fd) - 1)) {
        *max_fd = fd + 1;
      }
    }
  }
}

int shell_receive(fdmask)
struct fd_set *fdmask;
{
  int channel;
  int result;
  char buffer[BUFFER_LEN];
  int len;
  int status;
  pid_t pid;
  
  result = 1;
  /* now handle data */
  for (channel=0;channel<CHANNELS;channel++) {
    if (sh_stat[channel].active) {
      pty_check_timeout(channel);
#ifdef USE_SOCKET
      if ((sh_stat[channel].active == A_SOCKET) &&
          (!sh_stat[channel].connected)) {
        if (FD_ISSET(sh_stat[channel].sockfd,fdmask)) {
          if ((sh_stat[channel].pty = accept(sh_stat[channel].sockfd,
              (struct sockaddr *) &cli_addr, &clilen)) < 0) {
          }
          else {
            sh_stat[channel].connected = 1;
            result = 0;
          }
        }
        continue;
      }
#endif
      if (FD_ISSET(sh_stat[channel].pty,fdmask)) {
        if (read_data_pty(channel,buffer)) {
          result = 0;
        }
#ifdef READ_NONBLOCK
	result = 0;
#endif
      }
    }
  }
  /* clean up all remaining zombies */
  while ((pid = waitpid(-1,&status,WNOHANG)) > 0) {
    for (channel=0;channel<CHANNELS;channel++) {
      if (sh_stat[channel].active) {
        if (sh_stat[channel].pid == pid) {
          close_shell(channel,1);
        }
      }
    }
  }
  return (result);
}
#else
int shell_receive()
{
  int channel;
  int result;
  char buffer[BUFFER_LEN];
  int len;
  int nodata;
  int status;
  pid_t pid;
  
  result = 1;
  /* now handle data */
  for (channel=0;channel<CHANNELS;channel++) {
    nodata = 0;
    if (sh_stat[channel].active) {
      pty_check_timeout(channel);
    }
#ifdef USE_SOCKET
    if ((sh_stat[channel].active == A_SOCKET) &&
        (!sh_stat[channel].connected)) {
      if ((sh_stat[channel].pty = accept(sh_stat[channel].sockfd,
              (struct sockaddr *) &cli_addr, &clilen)) < 0) {
        nodata = 1;
      }
      else {
        fcntl(sh_stat[channel].pty,F_SETFL,O_NONBLOCK);
        sh_stat[channel].connected = 1;
        result = 0;
      }
      continue;
    }
#endif    
    while (sh_stat[channel].active && senddata_allowed(channel) && !nodata) {
      if (read_data_pty(channel,buffer)) {
        result = 0;
      }
      else {
        nodata = 1;
      }
    }
  }
  /* clean up all remaining zombies */
  while ((pid = waitpid(-1,&status,WNOHANG)) > 0) {
    for (channel=0;channel<CHANNELS;channel++) {
      if (sh_stat[channel].active) {
        if (sh_stat[channel].pid == pid) {
          close_shell(channel,1);
        }
      }
    }
  }
  return (result);
}
#endif

void init_shell()
{
  int i;
  
  for (i=0;i<CHANNELS;i++) {
    sh_stat[i].active = 0;
    sh_stat[i].pid = 0;
  }
  fixutmpfile();
}

void exit_shell()
{
  int i;
  
  for (i=0;i<CHANNELS;i++) {
    close_shell(i,0);
#ifdef USE_SOCKET
    close_socket(i,0);
#endif    
  }
}

#ifdef TNT_CHAMBER
/* open a shell on the current channel */
void cmd_frontend(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{

  char *env = 0;
  char slave[80];
  int i;
  char ucall[100],tmp[10];
  struct passwd *pw;
  struct termios termios;
  struct utmp utmpbuf;

  strip_call(&ucall,channel);
  if (sh_stat[channel].active) {
    cmd_display(mode,channel,"Shell or redirection already active",1);
    return;
  }
  if ((sh_stat[channel].pty = find_pty(&sh_stat[channel].num, slave)) < 0) {
    cmd_display(mode,channel,"Can't open pseudo terminal",1);
    return;
  }
  if(!strncmp(ucall,"DB0",3)) {
    cmd_display(mode,channel,"Uncorrect connect from Digipeater",1);
    disc_frontend(channel);
    return;
  }

  sh_stat[channel].active = A_FRONTEND;
  sh_stat[channel].mode = mode;
  sh_stat[channel].buflen = 0;
  strcpy(sh_stat[channel].id, slave + strlen(slave) - 2);
  if (!(sh_stat[channel].pid = fork())) {
/*
    pw = getpasswdentry(find_user_name(ch_stat[channel].call), 1);
    if (!pw || pw->pw_passwd[0]) pw = getpasswdentry(DEFAULTUSER, 0);
*/
/*
    for (i = 0; i < FD_SETSIZE; i++) close(i);
    setsid();
*/
    for (i = 0; i < _NFILE; i++) close(i);
    setpgrp();
    open(slave, O_RDWR, 0666);
    dup(0);
    dup(0);
    chmod(slave, 0622);
    memset((char *) &termios, 0, sizeof(termios));
    termios.c_iflag = ICRNL | IXOFF;
    termios.c_oflag = OPOST | TAB3 | ONLRET;
    termios.c_oflag &= ~(OCRNL | ONLCR | ONOCR);
    termios.c_cflag = CS8 | CREAD | CLOCAL;
    termios.c_lflag = ISIG | ICANON;
    termios.c_cc[VINTR]  = 127;
    termios.c_cc[VQUIT]  =  28;
    termios.c_cc[VERASE] =   8;
    termios.c_cc[VKILL]  =  24;
    termios.c_cc[VEOF]   =   4;
    cfsetispeed(&termios, B1200);
    cfsetospeed(&termios, B1200);
    tcsetattr(0, TCSANOW, &termios);
/*
    if (!pw || pw->pw_passwd[0]) exit(1);
*/
#ifdef LOGIN_PROCESS
    memset(&utmpbuf, 0, sizeof(utmpbuf));
    strcpy(utmpbuf.ut_user, "LOGIN");
    strcpy(utmpbuf.ut_id, sh_stat[channel].id);
    strcpy(utmpbuf.ut_line, slave + 5);
    utmpbuf.ut_pid = getpid();
    utmpbuf.ut_type = LOGIN_PROCESS;
    utmpbuf.ut_time = secclock(); 
#ifdef HAS_UTHOST
    strncpy(utmpbuf.ut_host,"AX25", sizeof(utmpbuf.ut_host));
#endif
    pututline(&utmpbuf);
    endutent();
#endif
    sprintf(tmp,"%d",channel);
    execl("/tcp/bin/sp-emu", "sp-emu", ucall, tmp, (char *) 0);
    exit(1);
  }
  cmd_display(mode,channel,"Ok",1);
  return;
}
#endif
